@@ -1,6 +1,7 @@
-from datetime import datetime
-from pprint import pformat
 import logging
+from pprint import pformat
+from datetime import datetime
+from contextlib import contextmanager
 
 from sqlalchemy import func
 from sqlalchemy.ext.declarative import declared_attr
@@ -18,6 +19,21 @@ else:
     logger = logging.getLogger("flask.app")
 
 
+@contextmanager
+def db_session_maker(auto_commit=False):
+    ## 不要尝试重新手动创建 session, eg: `db.create_scoped_session(options=options)`
+    ## 避免ORM对象在不同session中重复Attach
+    sess = db.session
+    yield sess
+    if auto_commit:
+        try:
+            sess.commit()
+        except Exception as e:
+            sess.rollback()
+            raise e
+    sess.close()
+
+
 def force_remove_multiple(*models, silent=False):
     count = 0
     for m in models:
@@ -97,17 +113,17 @@ class BaseModel():
 
     @classmethod
     def _make_query(cls, condition=None, limit=None, offset=None, order_by=None, **condition_kws):
-        # Can't call Query.update() or Query.delete() when limit()/offset()/distinct()/group_by()/order_by() has been called
+        # NOTE: ERROR raise if call query.[update({})/delete()] after limit()/offset()/distinct()/group_by()/order_by()
         condition = cls.strict_form(condition, **condition_kws)
         qry = cls.query.filter_by(**condition)
-        if order_by is not None:
-            if isinstance(order_by, (list, tuple)):
-                qry = qry.order_by(*order_by)
-            else:
-                qry = qry.order_by(order_by)
+        if isinstance(order_by, (list, tuple)):
+            qry = qry.order_by(*order_by)
+        elif order_by is not None:
+            # NOTE: Raise Error if bool(order_by)
+            qry = qry.order_by(order_by)
         if isinstance(limit, int):
             qry = qry.limit(limit)
-        if isinstance(offset, int):
+        if isinstance(offset, int) and offset >= 0:
             qry = qry.offset(offset)
         return qry
 
@@ -115,15 +131,31 @@ class BaseModel():
     def discard(cls, condition=None, limit=1, **condition_kws):
         # In Case of incorrect operation, default limit 1;
         condition = cls.strict_form(condition, **condition_kws)
-        n = cls.query.filter_by(**condition).delete()
-        if limit and limit < n:
-            db.session.rollback()
-            msg = "You're trying discard {} rows of {}, which is over limit={}".format(n, cls.__name__, limit)
-            raise errors.SecurityError(msg)
-        else:
-            db.session.commit()
+        with db_session_maker as db_sess:
+            # n = cls.query.filter_by(**condition).delete()
+            n = db_sess.query(cls).filter_by(**condition).delete()
+            if limit and limit < n:
+                db_sess.rollback()
+                msg = "You're trying discard {} rows of {}, which is over limit={}".format(n, cls.__name__, limit)
+                raise errors.SecurityError(msg)
+            else:
+                db.session.commit()
             return n
 
+    @classmethod
+    def page_items(cls, condition=None, limit=10, offset=0, order_by=None, **condition_kws):
+        qry = cls._make_query(condition, **condition_kws)
+        pk = cls.primary_keys()[0]
+        total = qry.value(func.count(getattr(cls, pk)))
+        if isinstance(order_by, (list, tuple)):
+            qry = qry.order_by(*order_by)
+        elif order_by:
+            qry = qry.order_by(order_by)
+        items = qry.limit(limit).offset(offset).all()
+        next_offset = offset + len(items)
+        has_more = total > next_offset
+        return dict(total=total, next_offset=next_offset, has_more=has_more, items=items)
+
     @classmethod
     def filter_by(cls, condition=None, **condition_kws):
         qry = cls._make_query(condition, **condition_kws)
@@ -198,6 +230,7 @@ class BaseModel():
 
 
 class CoModel(BaseModel):
+
     @declared_attr
     def created_time(self):
         return db.Column(db.DateTime, default=datetime.utcnow)
@@ -222,3 +255,16 @@ class CoModel(BaseModel):
             form['updated_time'] = parse_date(updated_time)
         m = cls(**form)
         return m
+
+    @classmethod
+    def lastOrNone(cls, **kwargs):
+        qry = cls._make_query(kwargs, limit=1, order_by=cls.created_time.desc())
+        return qry.one_or_none()
+
+    @classmethod
+    def lastOr404(cls, **kwargs):
+        m = cls.lastOrNone(**kwargs)
+        if isinstance(m, cls):
+            return m
+        msg = "Data Not Found: {}: {}".format(cls.__name__, pformat(kwargs))
+        raise errors.NotFound(msg)
